from abml_dataclass import Abml_Registry
from abml_helpers import Abml_Helpers, cprint, get_objects
from regionToolset import Region
from abaqusConstants import ON, OFF, COMPUTED, SPECIFIED
from abaqus import mdb, session


@Abml_Registry.register("constraints")
class Abml_Constraints(Abml_Helpers):
    def __init__(self, name, model, **kwargs):  # noqa
        self.name = name
        self.model = model
        self.kwargs = kwargs

        self.tolerance_map = {"computed": COMPUTED, "specified": SPECIFIED}

        self.create()

    def create(self):
        position_tolerance_method = self.kwargs.get("position_tolerance_method", "computed").lower()

        kwargs = {
            "name": self.name,
            "master": Region(side1Faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["master"])),
            "slave": Region(side1Faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["slave"])),
            "positionToleranceMethod": self.tolerance_map[position_tolerance_method],
            "adjust": self.kwargs.get("adjust", True),
            "tieRotations": self.kwargs.get("tie_rotations", True),
            "thickness": self.kwargs.get("thickness", True),
        }

        self.model.m.Tie(**kwargs)



from abml_dataclass import Abml_Registry

# from abml_helpers import cprint
from abaqus import mdb
from abaqusConstants import ANALYSIS, DEFAULT, OFF
import os
from shutil import copy, move
from io import open
from abml_helpers import cprint


@Abml_Registry.register("jobs")
class Abml_Job:
    def __init__(self, name, model, **kwargs):  # noqa
        self.name = str(name)
        self.model = model
        self.kwargs = kwargs
        self.create()

    def create(self):
        subroutine_flag = self.kwargs.get("subroutine", False)
        if subroutine_flag:
            subroutine_path = os.path.abspath("{}.for".format(self.name))
        else:
            subroutine_path = None

        kwargs = {
            "name": self.name,
            "model": self.model.name,
            "description": self.kwargs.get("description", ""),
            "type": ANALYSIS,
            "userSubroutine": subroutine_path,
            "numCpus": self.kwargs.get("cpus", 1),
            "numGPUs": self.kwargs.get("gpus", 0),
            "multiprocessingMode": DEFAULT,
        }

        mdb.Job(**kwargs)

        if self.kwargs.get("write_input", False):
            self.write_input()
            self.add_to_header(subroutine=self.sub_name, priority=self.kwargs.get("priority", "default"))

    def write_input(self):
        mdb.jobs[self.name].writeInput(consistencyChecking=OFF)

    def load_subname(self):
        if "subroutine_queue" in self.kwargs:
            self.sub_name = self.kwargs["subroutine_queue"]
        else:
            self.sub_name = "{}.for".format(self.name)

    def write_and_copy_input_to_path(self, path):
        self.write_input()
        self.add_to_header(subroutine="{}.for".format(self.name), priority=self.kwargs.get("priority", "default"))
        filename = "{}.inp".format(self.name)
        copy(filename, os.path.join(path, filename))

    def write_and_move_input_to_path(self, path):
        self.write_input()
        self.add_to_header(subroutine="{}.inp".format(self.name), priority=self.kwargs.get("priority", "default"))
        filename = "{}.inp".format(self.name)
        move(filename, os.path.join(path, filename))

    def add_to_header(self, **kwargs):
        filename = "{}.inp".format(self.name)
        with open(filename, "r", encoding="utf-8") as f:
            input_file = f.readlines()

        for key, value in kwargs.items():
            input_file.insert(1, "** {key}={value}\n".format(key=key, value=value))

        with open(filename, "w", encoding="utf-8") as f:
            input_file = "".join(input_file)
            f.write(input_file)



import types
import json
import inspect
from collections import OrderedDict
from collections import defaultdict
import sys
import yaml
from io import open
from abaqus import mdb
from abaqusConstants import STANDARD_EXPLICIT
from abml import abml_helpers
import interaction
from abaqus import openMdb


class Abml_Registry:
    registry = defaultdict(lambda: OrderedDict)

    @classmethod
    def register(cls, key):
        def decorator(dataclass):
            cls.registry[key] = dataclass
            return dataclass

        return decorator


@Abml_Registry.register("cae")
class Abml_Cae:
    def __init__(self, models, **kwargs):
        self.kwargs = kwargs

        cae_path = self.kwargs.get("cae", None)
        cae_path = models.pop("cae") if "cae" in models else None
        if cae_path is not None:
            self.open_cae(cae_path)

        self.models = to_object_key_mode(models["models"], type_="models")

    def open_cae(self, path):
        openMdb(pathName=path)

    def save_cae(self, filename):
        mdb.saveAs(pathName=filename)


@Abml_Registry.register("models")
class Abml_Model:
    """
    Creates an Abaqus/Standard model using the provided input.

    Parameters
    ----------
    name : str
        The name of the Abaqus/Standard model.
    **kwargs : dict
        A dictionary of inputs for creating the model. The dictionary can contain the following keys:

        - materials : dict, optional
            A dictionary of materials to be added to the model.
        - sections : dict, optional
            A dictionary of sections to be added to the model.
        - parts : dict, optional
            A dictionary of parts to be added to the model.
        - assembly : dict, optional
            A dictionary describing the assembly of parts in the model.
        - steps : dict, optional
            A dictionary of steps to be added to the model.
        - interaction_props : dict, optional
            A dictionary of interaction properties to be added to the model.
        - interactions : dict, optional
            A dictionary of interactions to be added to the model.
        - constraints : dict, optional
            A dictionary of constraints to be added to the model.
        - bcs : dict, optional
            A dictionary of boundary conditions to be added to the model.
        - loads : dict, optional
            A dictionary of loads to be added to the model.
        - jobs : dict, optional
            A dictionary of jobs to be added to the model.
        - modules : dict, optional
            A dictionary of sub-modules to be created within the model.

        Methods
        -------
        create()
            Creates an instance of a model in the cae
        create_instances(**kwargs)
            Creates instances of materials, sections, parts, etc. in the model.
        create_materials(**kwargs)
            creates material
        create_sections(**kwargs)
            creates sections
        create_parts(**kwargs)
            creates material
        create_assembly(**kwargs)
            creates the assembly
        create_steps(**kwargs)
            creates steps
        create_interaction_props(**kwargs)
            creates interaction props
        create_interactions(**kwargs)
            creates interactionss
        create_constraints(**kwargs)
            creates constraints
        create_bcs(**kwargs)
            creates bcs
        create_loads(**kwargs)
            creates loads
        create_jobs(**kwargs)
            creates jobs
        create_modules(**kwargs)
            creates modules


        Returns
        -------
        None
    """

    def __init__(self, name, **kwargs):
        self.name = name

        self.create()
        self.m, self.a, self.p = abml_helpers.get_objects(self.name)
        self.kwargs = kwargs

        self.materials = {}
        self.sections = {}
        self.parts = {}
        self.assembly = {}
        self.steps = {}
        self.interaction_props = {}
        self.interactions = {}
        self.constraints = {}
        self.bcs = {}
        self.loads = {}
        self.jobs = {}
        self.sketches = {}

        self.create_instances(**kwargs)

    def create_instances(self, **kwargs):
        create_order = [
            self.create_sketches,
            self.create_materials,
            self.create_sections,
            self.create_parts,
            self.create_assembly,
            self.create_steps,
            self.create_interaction_props,
            self.create_interactions,
            self.create_constraints,
            self.create_bcs,
            self.create_loads,
            self.create_jobs,
            self.create_modules,
        ]

        for creation in create_order:
            creation(**kwargs)

    def update_nested_dict(self, nested_dict, update_dict):
        """
        Recursively update a nested dictionary with a dictionary of key-value pairs
        if the keys are not already present.
        """
        for k, v in update_dict.iteritems():
            if isinstance(v, dict):
                # Recursively update nested dictionaries
                nested_dict[k] = self.update_nested_dict(nested_dict.get(k, {}), v)
            elif k not in nested_dict:
                # Key is not present, update the dictionary
                nested_dict[k] = v
        return nested_dict

    def create_sketches(self, **kwargs):
        sketches_list = [value for key, value in kwargs.items() if "sketches" in key]
        for sketches in sketches_list:
            if sketches is not None:
                sketches = OrderedDict(sorted(sketches.items()))

                sketch_name = sketches.keys()[0]
                sketch_cmds = sketches.values()[0]

                self.sketches.update(
                    {sketch_name: to_object_cmd_mode(sketch_cmds, "sketch", model=self, name=sketch_name)}
                )

    def create_materials(self, **kwargs):
        materials_list = [value for key, value in kwargs.items() if "materials" in key]
        for materials in materials_list:
            if materials is not None:
                materials = OrderedDict(sorted(materials.items()))
                self.materials.update(to_object_key_mode(materials, "materials", model=self))

    def create_sections(self, **kwargs):
        sections_list = [value for key, value in kwargs.items() if "sections" in key]
        for sections in sections_list:
            if sections is not None:
                sections = OrderedDict(sorted(sections.items()))
                self.sections.update(to_object_key_mode(sections, "sections", model=self))

    def create_parts(self, **kwargs):
        parts_list = [value for key, value in kwargs.items() if "parts" in key]
        for parts in parts_list:
            if parts is not None:
                parts = OrderedDict(sorted(parts.items()))
                self.parts.update(to_object_key_mode(parts, "parts", model=self))

    def create_assembly(self, **kwargs):
        assembly_list = [value for key, value in kwargs.items() if "assembly" in key]
        for assembly in assembly_list:
            if assembly is not None:
                self.assembly = to_object_hierachy_mode(assembly, "assembly", model=self)

    def create_steps(self, **kwargs):
        steps_list = [value for key, value in kwargs.items() if "steps" in key]
        for steps in steps_list:
            if steps is not None:
                steps = OrderedDict(sorted(steps.items()))
                self.interaction_props.update(to_object_key_mode(steps, "steps", model=self))

    def create_interaction_props(self, **kwargs):
        interaction_props_list = [value for key, value in self.kwargs.items() if "interaction_props" in key]
        for interaction_props in interaction_props_list:
            if interaction_props is not None:
                interaction_props = OrderedDict(sorted(interaction_props.items()))
                self.interaction_props.update(to_object_key_mode(interaction_props, "interaction_props", model=self))

    def create_interactions(self, **kwargs):
        interactions_list = [value for key, value in kwargs.items() if "interactions" in key]
        for interactions in interactions_list:
            if interactions is not None:
                interactions = OrderedDict(sorted(interactions.items()))
                self.interactions.update(to_object_key_mode(interactions, "interactions", model=self))

    def create_constraints(self, **kwargs):
        constraints_list = [value for key, value in kwargs.items() if "constraints" in key]
        for constraints in constraints_list:
            if constraints is not None:
                constraints = OrderedDict(sorted(constraints.items()))
                self.constraints.update(to_object_key_mode(constraints, "constraints", model=self))

    def create_bcs(self, **kwargs):
        bcs_list = [value for key, value in kwargs.items() if "bcs" in key]
        for bcs in bcs_list:
            if bcs is not None:
                bcs = OrderedDict(sorted(bcs.items()))
                self.bcs.update(to_object_key_mode(bcs, "bcs", model=self))

    def create_loads(self, **kwargs):
        loads_list = [value for key, value in kwargs.items() if "loads" in key]
        for loads in loads_list:
            if loads is not None:
                loads = OrderedDict(sorted(loads.items()))
                self.loads.update(to_object_key_mode(loads, "loads", model=self))

    def create_jobs(self, **kwargs):
        jobs_list = [value for key, value in kwargs.items() if "jobs" in key]
        for jobs in jobs_list:
            if jobs is not None:
                jobs = OrderedDict(sorted(jobs.items()))
                self.jobs.update(to_object_key_mode(jobs, "jobs", model=self))

    def create_modules(self, **kwargs):
        modules_list = kwargs.get("modules", [])
        # abml_helpers.cprint("############## {}".format(modules_list[0]))
        modules_list = [value for key, value in kwargs.items() if "modules" in key]
        # abml_helpers.cprint(modules_list[0])
        for modules in modules_list:
            if modules is not None:
                if isinstance(modules, dict):
                    modules = OrderedDict(sorted(modules.items()))
                    self.create_instances(**modules)
                elif isinstance(modules, list):
                    for module in modules:
                        module = OrderedDict(sorted(module.items()))
                        self.create_instances(**module)

    def create(self):
        if self.name not in mdb.models.keys():
            mdb.Model(modelType=STANDARD_EXPLICIT, name=self.name)

        if "Model-1" in mdb.models.keys():
            del mdb.models["Model-1"]


def data_to_object(data, type_="models", key_based=True, **kwargs):
    """
    Converts input data into objects of specified type.

    Parameters
    ----------
    data : dict, list, or None
        Input data to be converted into objects. If `None`, returns `None`.
    type_ : str, optional
        Type of objects to be created. Defaults to "models".
    key_based : bool, optional
        Whether to create objects based on the keys in `data` (if `True`) or create a single object
        from `data` (if `False`). Defaults to `True`.
    **kwargs
        Additional keyword arguments to pass to object constructors.

    Returns
    -------
    object_
        If `data` is a dictionary, returns a dictionary of objects, where each key in `data`
        corresponds to an object created from the corresponding value. If `key_based` is `False`
        and `data` is a dictionary, returns a single object created from the values in `data`.
        If `data` is a list, returns a list of objects created from the elements in `data`.
        If `data` is `None`, returns `None`.

    """
    if data is not None:
        object_ = {}
        if isinstance(data, dict):
            object_ = {}
            if key_based:
                for key, value in data.items():
                    if isinstance(value, dict):
                        value.update(kwargs)
                        object_[key] = Abml_Registry.registry[type_](name=key, **value)
                    elif isinstance(value, (list)):
                        object_[key] = Abml_Registry.registry[type_](name=key, *value)
            else:
                data.update(kwargs)
                object_ = Abml_Registry.registry[type_](**data)
        elif isinstance(data, list):
            object_ = []
            for value in data:
                value.update(kwargs)
                object_.append(Abml_Registry.registry[type_](**value))

        return object_
    return None


def to_object_key_mode(data, type_, **kwargs):
    """
    Create objects of a specified type from a dictionary of data, where each key in the dictionary represents an object to be created.

    Parameters
    ----------
    data : dict
        A dictionary where each key represents an object to be created, and each value contains
        the data necessary to create the object.
    type_ : str
        The type of object to be created. This must correspond to a valid key in the `registry`
        dictionary of the `Abml_Registry` class.
    **kwargs : dict, optional
        Additional keyword arguments to be passed to the object constructor.

    Returns
    -------
    dict
        A dictionary where each key is the name of the object, and each value is the object itself.

    """
    object_ = {}
    for key, value in data.items():
        value.update(kwargs)
        object_[key] = Abml_Registry.registry[type_](name=key, **value)

    return object_


def to_object_option_mode(data, type_, **kwargs):
    """
    Convert a dictionary of data to an object of the specified type.

    Parameters
    ----------
    data : dict
        A dictionary containing the data to convert to an object.
    type_ : str
        The type of object to create.
    **kwargs : dict
        Additional keyword arguments to pass to the object constructor.

    Returns
    -------
    object_
        An object of the specified type containing the data from the dictionary.
    """
    data.update(kwargs)
    object_ = Abml_Registry.registry[type_](**data)

    return object_


def to_object_key_based_type(data, **kwargs):
    """
    Converts a dictionary of data into a collection of objects based on the key.

    Parameters
    ----------
    data : dict
        The dictionary of data to be converted into objects.
    **kwargs : keyword arguments, optional
        Additional arguments to be passed to the object constructor.

    Returns
    -------
    object_ : dict
        A dictionary of objects created from the data, where each object is associated with its corresponding key in the input dictionary.

    Notes
    -----
    This function assumes that each key in the input dictionary corresponds to the name of a class in the `Abml_Registry` registry. It creates an instance of each class with the corresponding data value as constructor arguments, and returns a dictionary where each object is associated with its corresponding key.

    """
    object_ = {}
    for key, value in data.items():
        value.update(kwargs)
        object_[key] = Abml_Registry.registry[key](**value)

    return object_


def to_object_cmd_mode(data, type_, **kwargs):
    """
    Convert a list of dictionaries representing command data into a list of objects.

    Parameters
    ----------
    data : list of dict
        List of dictionaries containing command data.
    type_ : str
        The type of object to create from the command data.
    **kwargs : dict, optional
        Optional keyword arguments to be passed to the object constructor.

    Returns
    -------
    list of object
        A list of objects created from the command data.

    Notes
    -----
    This function expects each dictionary in the `data` list to have a single key-value pair,
    where the key is the command name and the value is a dictionary containing the command arguments.
    The `type_` parameter specifies the type of object to create from the command data, and must be a
    string that matches the name of a class registered with the `Abml_Registry.registry` dictionary.

    Examples
    --------
    **python
    >>> data = [{'command1': {'arg1': 1, 'arg2': 2}}, {'command2': {'arg1': 3, 'arg2': 4}}]
    >>> objects = to_object_cmd_mode(data, 'MyClass', arg3=5)
    >>> objects[0].arg1
    1
    >>> objects[1].arg2
    4
    """

    object_ = []
    for cmd in data:
        value = cmd.values()[0]
        key = cmd.keys()[0]
        value.update(kwargs)
        object_.append(Abml_Registry.registry[type_](cmd=key, **value))

    return object_


def to_object_hierachy_mode(data, type_, **kwargs):
    """
    Create an object from a hierarchical dictionary representation of the object's attributes.

    Parameters
    ----------
    data : dict
        A hierarchical dictionary representation of the object's attributes.
    type_ : str
        The name of the class to create the object from.
    **kwargs : dict, optional
        Optional keyword arguments to be passed to the object constructor.

    Returns
    -------
    object
        An instance of the specified class, with attributes set according to the data dictionary.

    Notes
    -----
    This function expects the `data` dictionary to be hierarchical, with each key representing an attribute
    name and each value representing the attribute value. If a value is itself a dictionary, the attribute
    is treated as a sub-object, and this function is recursively called to create the sub-object. The `type_`
    parameter specifies the name of the class to create the object from, and must be a string that matches the
    name of a class registered with the `Abml_Registry.registry` dictionary.

    """
    data.update(kwargs)
    object_ = Abml_Registry.registry[type_](**data)

    return object_



from abml_dataclass import Abml_Registry
from abml_helpers import Abml_Helpers, cprint, get_kwargs_string
from abaqusConstants import USER_DEFINED, UNSET, UNIFORM
from regionToolset import Region
import logging

logger = logging.getLogger("abml_logger")


@Abml_Registry.register("loads")
class Abml_Loads(Abml_Helpers):
    def __init__(self, name, model, type, **kwargs):  # noqa
        self.name = name
        self.model = model
        self.type = type
        self.kwargs = kwargs
        self.map = {"pressure": self.pressure, "gravity": self.gravity}
        self.distribution_type_map = {
            "user_defined": USER_DEFINED,
            "uniform": UNIFORM,
        }

        logger.debug(get_kwargs_string(model=model.name, name=name, type=type))

        self.create()

    def create(self):
        self.map[self.type]()

    def pressure(self):
        logger.debug(get_kwargs_string(type="pressure", faces=self.kwargs["faces"]))
        kwargs = {
            "name": self.name,
            "createStepName": self.kwargs["step"],
            "region": Region(side1Faces=self.model.assembly.get_seq_from_surface_list(self.kwargs["faces"])),
            "distributionType": self.distribution_type_map[self.kwargs.get("distribution_type", "uniform")],
            "magnitude": self.kwargs.get("magnitude", 0.0),
            "amplitude": self.kwargs.get("amplitude", UNSET),
        }

        self.model.m.Pressure(**kwargs)

    def gravity(self):
        kwargs = {
            "name": self.name,
            "createStepName": self.kwargs["step"],
            "region": None,
            "distributionType": self.distribution_type_map[self.kwargs.get("distribution_type", "uniform")],
            "comp1": self.kwargs.get("comp1", 0.0),
            "comp2": self.kwargs.get("comp2", 0.0),
            "comp3": self.kwargs.get("comp3", 0.0),
            "amplitude": self.kwargs.get("amplitude", UNSET),
        }

        self.model.m.Gravity(**kwargs)



import abml_helpers
from abaqusConstants import THREE_D, DEFORMABLE_BODY, XZPLANE  # noqa
from abaqusConstants import XYPLANE, YZPLANE, XZPLANE  # noqa
from abaqusConstants import MIDDLE_SURFACE, FROM_SECTION  # noqa
from abaqusConstants import ON, OFF, STANDALONE  # noqa
from regionToolset import Region
from abml_dataclass import Abml_Registry
from abml_dataclass import to_object_option_mode, to_object_key_mode, to_object_cmd_mode
from abml_helpers import Abml_Helpers
from abml_sketch import Abml_Sketch
from numpy import array, abs
from abaqus import mdb


@Abml_Registry.register("parts")
class Abml_Part:
    """
    A class representing a part in an Abaqus/Explicit model.

    Parameters
    ----------
    name : str
        The name of the part.
    model : :ref:`Abml_Model <Abml_Model>`
        The parent model to which the part belongs.
    **kwargs
        Additional keyword arguments that specify the geometry, planes, partitions,
        surfaces, sets, section assignments, and mesh associated with the part.

    Attributes
    ----------
    model : :ref:`Abml_Model <Abml_Model>`
        The parent model to which the part belongs.
    name : str
        The name of the part.
    geometry : abaqus.geom.GeometryObject or None
        The geometry object associated with the part, if any.
    planes : abaqus.core.AbaqusDict or None
        The planes associated with the part, if any.
    partitions : abaqus.core.AbaqusDict or None
        The partitions associated with the part, if any.
    surfaces : abaqus.core.AbaqusDict or None
        The surfaces associated with the part, if any.
    sets : abaqus.core.AbaqusDict or None
        The sets associated with the part, if any.
    section_assignments : abaqus.core.AbaqusDict or None
        The section assignments associated with the part, if any.
    mesh : abaqus.core.AbaqusDict or None
        The mesh associated with the part, if any.

    Methods
    -------
    create()
        Creates the part in the model.
    get_sequence_from_set_list(set_list)
        Returns a dictionary of sequences of faces, cells, and edges for the specified sets.
    generate_mesh()
        Generates the mesh for the part.

    """

    def __init__(self, name, model, **kwargs):  # noqa
        self.model = model
        self.name = name
        self.create()

        if self.name in model.parts:
            self.inherit(model.parts[self.name])

        geometry = kwargs.get("geometry")
        if geometry is not None:
            params = {
                "data": geometry,
                "type_": "geometry",
                "part": self,
            }
            self.geometry = to_object_option_mode(**params)

        planes = kwargs.get("planes")
        if planes is not None:
            params = {
                "data": planes,
                "type_": "planes",
                "part": self,
            }
            self.planes = to_object_key_mode(**params)

        partitions = kwargs.get("partitions")
        if partitions is not None:
            params = {
                "data": partitions,
                "type_": "partitions",
                "part": self,
            }
            self.partitions = to_object_cmd_mode(**params)

        surfaces = kwargs.get("surfaces")
        if surfaces is not None:
            params = {
                "data": surfaces,
                "type_": "surfaces",
                "part": self,
            }
            self.surfaces = to_object_key_mode(**params)
            # to_object_key_mode(**params)

        sets = kwargs.get("sets")

        if sets is not None:
            params = {
                "data": sets,
                "type_": "sets",
                "part": self,
            }
            self.sets = to_object_key_mode(**params)

        section_assignments = kwargs.get("section_assignments")
        if section_assignments is not None:
            params = {
                "data": section_assignments,
                "type_": "section_assignments",
                "part": self,
            }
            self.section_assignments = to_object_key_mode(**params)

        mesh = kwargs.get("mesh")
        if mesh is not None:
            params = {
                "data": mesh,
                "type_": "mesh",
                "part": self,
            }

            self.mesh = to_object_cmd_mode(**params)
            self.generate_mesh()

    def inherit(self, other):
        if isinstance(other, Abml_Part):
            for attr in dir(other):
                if not attr.startswith("__") and not callable(getattr(other, attr)):
                    setattr(self, attr, getattr(other, attr))

    def create(self):
        if self.name not in self.model.parts.keys():
            self.model.m.Part(dimensionality=THREE_D, name=self.name, type=DEFORMABLE_BODY)

    def get_sequence_from_set_list(self, set_list):
        face_seq = None
        cell_seq = None
        edge_seq = None
        for set_ in set_list:
            seq = self.model.p[self.name].sets[set_].faces
            if face_seq is None:
                face_seq = seq
            else:
                face_seq += seq

            seq = self.model.p[self.name].sets[set_].cells
            if cell_seq is None:
                cell_seq = seq
            else:
                cell_seq += seq

            seq = self.model.p[self.name].sets[set_].edges
            if edge_seq is None:
                edge_seq = seq
            else:
                edge_seq += seq

        kwargs = {
            "faces": face_seq,
            "cells": cell_seq,
            "edges": edge_seq,
        }

        return kwargs

    def generate_mesh(self):
        self.model.p[self.name].generateMesh()


@Abml_Registry.register("geometry")
class Abml_Geometry(Abml_Helpers):
    """
    A class for creating geometries in Abaqus/CAE.

    Parameters
    ----------
    type : str
        | The type of geometry to create. Possible values are
        | solid_extrusion
        | solid_revolve
    sketch : dict
        A dictionary containing the data for the sketch to create the geometry.
    part : Part object
        The part in which to create the geometry.
    **kwargs
        Additional keyword arguments to pass to the geometry creation function.

    Methods
    -------
    create()
        Creates the geometry using the appropriate creation function based on the specified type.

    Notes
    -----
    This class is intended to be used with the Abaqus finite element analysis software.

    Examples
    --------
    part_1:
      geometry:
        type: solid_extrusion
        sketch:
          - rect:
            p1: [0.0, 0.0]
            p2: [1.0, 1.0]
        depth: 1.0
    """

    def __init__(self, type, part, **kwargs):  # noqa
        self.type = type
        self.part = part
        self.model = part.model
        self.kwargs = kwargs
        self.sketch_name = self.kwargs.get("sketch_name", "__profile__")
        params = {"data": kwargs.get("sketch", []), "type_": "sketch", "model": self.model, "name": self.sketch_name}
        self.sketch = to_object_cmd_mode(**params)

        self.map = {
            ("3d", "deformable", "solid_extrusion"): self._create_solid_extrusion,
            ("3d", "deformable", "solid_revolve"): self._create_solid_revolve,
        }

        self.create()

    def create(self):
        self.map[("3d", "deformable", self.type)]()

    def _create_solid_extrusion(self):
        if not isinstance(self.kwargs["depth"], (int, float)) or self.kwargs["depth"] == 0.0:
            error_str = "Extrusion depth for part: {} is no int, float or is 0".format(self.part.name)
            raise ValueError(error_str)
        part = self.model.p[self.part.name]
        sketch = self.model.m.sketches[self.sketch_name]
        part.BaseSolidExtrude(depth=self.kwargs["depth"], sketch=sketch)

        if "__profile__" in self.model.m.sketches:
            del self.model.m.sketches["__profile__"]

        self.aliases = {}

    def _create_solid_revolve(self):
        self.sketch[0].create_construction_line()
        if not isinstance(self.kwargs["angle"], (int, float)) or self.kwargs["angle"] == 0.0:
            error_str = "Revolve angle for part: {} is no int, float or is 0".format(self.part.name)
            raise ValueError(error_str)
        part = self.model.p[self.part.name]
        sketch = self.model.m.sketches[self.sketch_name]

        part.BaseSolidRevolve(angle=self.kwargs["angle"], sketch=sketch, flipRevolveDirection=OFF)
        del mdb.models["m1"].sketches[self.sketch_name]

        if "__profile__" in self.model.m.sketches:
            del self.model.m.sketches["__profile__"]

        self.aliases = {}

    def get_face_sequence(self, element):
        part = self.model.p[self.part.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            face = part.faces.findAt(element.astype(float))  # noqa
            if face is not None:
                seq = Region(side1Faces=part.faces[face.index : face.index + 1]).side1Faces
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            faces = part.faces.getByBoundingBox(*element.astype(float))
            if len(faces) != 0:
                seq = faces
        elif isinstance(element.item(0), str):  # getByBoundingBox str
            if any(item == element.item(0) for item in ["all"]):
                seq = part.faces[:]
            elif any(item == element.item(0) for item in part.surfaces.keys()):
                seq = part.surfaces[element.item(0)].faces
            else:
                boundaries = self.get_boundary_rect(self.sketch[0], self.kwargs["depth"], element.item(0))
                seq = part.faces.getByBoundingBox(*boundaries)

        return seq

    def get_cell_sequence(self, element):
        part = self.model.p[self.part.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            cell = part.cells.findAt(element.astype(float))
            if cell is not None:
                seq = Region(cells=part.cells[cell.index : cell.index + 1]).cells
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            cells = part.cells.getByBoundingBox(*element.astype(float))
            if len(cells) != 0:
                seq = cells
        elif isinstance(element.item(0), str):  # getByBoundingBox
            if element.item(0) == "all":
                seq = part.cells[:]
            else:
                boundaries = self.get_boundary_rect(self.sketch[0], self.kwargs["depth"], element.item(0))
                seq = part.cells.getByBoundingBox(*boundaries)
        return seq

    def get_edge_sequence(self, element):
        part = self.model.p[self.part.name]
        element = array([element]).flatten()
        seq = None
        if element.shape == (3,) and isinstance(element[0], (float, int)):  # findAt
            edge = part.edges.findAt(element.astype(float))
            if isinstance(edge, type(None)):
                seq = None
            else:
                seq = Region(edges=part.edges[edge.index : edge.index + 1]).edges
        elif element.shape == (6,) and isinstance(element[0], (float, int)):  # getByBoundingBox
            edges = part.edges.getByBoundingBox(*element.astype(float))
            if len(edges) != 0:
                seq = edges
        elif isinstance(element.item(0), str):  # getByBoundingBox
            if any(item == element.item(0) for item in ["all"]):
                seq = part.edges[:]
            else:
                boundaries = self.get_boundary_rect(self.sketch[0], self.kwargs["depth"], element.item(0))
                seq = part.edges.getByBoundingBox(*boundaries)

        return seq

    def get_set_sequence_edges(self, element):
        return Region(edges=self.model.p[self.part.name].sets[element].edges).edges

    def get_set_sequence_cells(self, element):
        return Region(cells=self.model.p[self.part.name].sets[element].cells).cells

    def get_set_sequence_faces(self, element):
        return Region(side1Faces=self.model.p[self.part.name].sets[element].faces).faces

    def get_sequence_from_list(self, elements, type_):
        sequence_map = {
            "cell": self.get_cell_sequence,
            "face": self.get_face_sequence,
            "edge": self.get_edge_sequence,
            "set_edges": self.get_set_sequence_edges,
            "set_faces": self.get_set_sequence_faces,
            "set_cells": self.get_set_sequence_cells,
        }

        type_seq = None
        for element in elements:
            if len(element) != 0:
                seq = sequence_map[type_](element)
                if type_seq is None:
                    type_seq = seq
                else:
                    type_seq += seq
        return type_seq


@Abml_Registry.register("planes")
class Abml_Plane:
    """
    A class to create datum planes in Abaqus.

    Parameters
    ----------
    name : str
        The name of the datum plane to be created.
    part : Abml_Part
        The Abml_Part object to which the datum plane belongs.
    **kwargs : dict
        Additional keyword arguments to be passed to the Abaqus API. The keys of the dictionary should be the names of
        the principal planes ('x', 'y', or 'z'), and the values should be the offsets of the datum planes from the
        origin in the corresponding directions.

    Attributes
    ----------
    name : str
        The name of the datum plane.
    part : Abml_Part
        The Abml_Part object to which the datum plane belongs.
    model : Abml_Model
        The Abml_Model object to which the datum plane belongs.
    kwargs : dict
        The keyword arguments passed to the Abaqus API.
    _principal_Plane : dict
        A dictionary mapping the names of the principal planes ('x', 'y', or 'z') to the corresponding constants in
        the Abaqus API.

    Methods
    -------
    create()
        Create the datum plane in the Abaqus model.

    Notes
    -----
    This class uses the Abaqus API to create datum planes in Abaqus. The datum planes are created in the part
    specified by the `part` parameter. The position and orientation of the datum planes are specified by the
    `**kwargs` parameter. The keys of the `**kwargs` parameter should be the names of the principal planes ('x', 'y',
    or 'z'), and the values should be the offsets of the datum planes from the origin in the corresponding directions.
    """

    def __init__(self, name, part, **kwargs):  # noqa
        self.name = name
        self.part = part
        self.model = part.model
        self.kwargs = kwargs

        self._principal_Plane = {"x": YZPLANE, "y": XZPLANE, "z": XYPLANE}
        self.create()

    def create(self):
        part = self.model.p[self.part.name]
        for key, value in self.kwargs.items():
            part.DatumPlaneByPrincipalPlane(offset=value, principalPlane=self._principal_Plane[key])
            part.features.changeKey(fromName="Datum plane-1", toName=self.name)  # noqa


@Abml_Registry.register("partitions")
class Abml_Partition(Abml_Helpers):
    """
    A class for partitioning parts in Abaqus based on faces, edges, and cells.

    Parameters
    ----------
    part : abaqus Part object
        The part to be partitioned.
    faces : list of ints, optional
        A list of face IDs to be partitioned. Defaults to None.
    edges : list of ints, optional
        A list of edge IDs to be partitioned. Defaults to None.
    cells : list of ints, optional
        A list of cell IDs to be partitioned. Defaults to None.
    **kwargs : dict
        Optional keyword arguments to be passed to the partitioning function. The supported
        keys are 'cmd' (str) and 'planes' (list of str). 'cmd' specifies the partitioning
        method to use, and 'planes' is a list of datum plane feature IDs.

    Attributes
    ----------
    method : str
        The partitioning method to use.
    part : abaqus Part object
        The part to be partitioned.
    model : abaqus Model object
        The model containing the part to be partitioned.
    faces : list of lists of ints or str
        A list of face IDs to be partitioned. If None, no faces are partitioned.
    edges : list of lists of ints or str
        A list of edge IDs to be partitioned. If None, no edges are partitioned.
    cells : list of lists of ints or str
        A list of cell IDs to be partitioned. If None, no cells are partitioned.
    kwargs : dict
        Optional keyword arguments passed to the partitioning function.
    partition_map : dict
        A dictionary containing the partitioning method names and their corresponding
        partitioning functions.

    Methods
    -------
    create()
        Calls the partitioning function specified by 'method'.
    create_face_plane()
        Partitions the faces of the part by a datum plane.

    """

    def __init__(self, part, faces=None, edges=None, cells=None, **kwargs):
        self.method = kwargs.get("cmd", "plane")
        self.part = part
        self.model = part.model
        self.faces = self.convert_position_list(faces)
        self.edges = self.convert_position_list(edges)
        self.cells = self.convert_position_list(cells)
        self.kwargs = kwargs
        self.partition_map = {"plane": self.create_face_plane}

        self.create()

    def create(self):
        self.partition_map[self.method]()

    def create_face_plane(self):
        part_ = self.model.p[self.part.name]
        planes = array([self.kwargs.get("planes", [])]).flatten()

        for plane in planes:
            id_ = part_.features[plane].id

            if self.faces is not None:
                face_seq = self.part.geometry.get_sequence_from_list(self.faces, "face")
                if face_seq is not None:
                    try:
                        part_.PartitionFaceByDatumPlane(datumPlane=part_.datums[id_], faces=face_seq)
                    except Exception as e:
                        error_msg = """
                        part_name: {part_name}
                        number of face seq found: {num_face_seq}
                        plane: {plane}

                        {e}
                        """.format(
                            part_name=self.part.name,
                            e=e,
                            num_face_seq=len(face_seq),
                            plane=plane,
                        )
                        raise ValueError(error_msg)

            if self.cells is not None:
                cell_seq = self.part.geometry.get_sequence_from_list(self.cells, "cell")
                if cell_seq is not None:
                    part_.PartitionCellByDatumPlane(datumPlane=part_.datums[id_], cells=cell_seq)

            if self.edges is not None:
                edge_seq = self.part.geometry.get_sequence_from_list(self.edges, "edge")
                if edge_seq is not None:
                    part_.PartitionEdgeByDatumPlane(datumPlane=part_.datums[id_], edges=edge_seq)


@Abml_Registry.register("surfaces")
class Abml_Surface(Abml_Helpers):
    def __init__(self, name, faces, part, **kwargs):
        self.name = name
        self.part = part
        self.model = self.part.model
        self.faces = self.convert_position_list(faces)

        self.create()

    def create(self):
        part = self.model.p[self.part.name]

        try:
            face_seq = self.part.geometry.get_sequence_from_list(self.faces, "face")
        except AttributeError:
            face_seq = self.part.model.parts[self.part.name].geometry.get_sequence_from_list(self.faces, "face")

        try:
            part.Surface(side1Faces=face_seq, name=self.name)
        except Exception as e:
            error_msg = """
            surface_name: {name}
            faces_inp: {faces}
            face_seq: {face_seq}
            part: {part}
            {e}
            """.format(
                name=self.name,
                faces=self.faces,
                face_seq=face_seq,
                part=self.part.name,
                e=e,
            )
            raise ValueError(error_msg)


@Abml_Registry.register("sets")
class Abml_Sets(Abml_Helpers):
    def __init__(self, name, part, faces=[], edges=[], cells=[]):
        self.name = name
        self.part = part
        self.model = part.model
        self.faces = self.convert_position_list(faces)
        self.edges = self.convert_position_list(edges)
        self.cells = self.convert_position_list(cells)
        self.create()

    def create(self):
        part = self.model.p[self.part.name]

        kwargs = {}
        face_seq = self.part.geometry.get_sequence_from_list(self.faces, "face")
        if face_seq is not None:
            kwargs["faces"] = face_seq

        cell_seq = self.part.geometry.get_sequence_from_list(self.cells, "cell")
        if cell_seq is not None:
            kwargs["cells"] = cell_seq

        edge_seq = self.part.geometry.get_sequence_from_list(self.edges, "edge")
        if edge_seq is not None:
            kwargs["edges"] = edge_seq

        part.Set(name=self.name, **kwargs)


@Abml_Registry.register("section_assignments")
class Abml_Section_Assignments(Abml_Helpers):
    def __init__(self, name, part, cells=None, sets=None):
        self.name = name
        self.part = part
        self.model = part.model

        self.cells = self.convert_position_list(cells)
        self.sets = self.convert_position_list(sets)
        self.create()

    def create(self):
        cell_seq = self.part.geometry.get_sequence_from_list(self.cells, "cell")
        if self.sets is not None:
            set_seq = self.part.get_sequence_from_set_list(self.sets)["cells"]
            if set_seq is not None:
                cell_seq += set_seq

        kwargs = {
            "sectionName": self.name,
            "offset": 0.0,
            "offsetType": MIDDLE_SURFACE,
            "offsetField": "",
            "thicknessAssignment": FROM_SECTION,
            "region": Region(cells=cell_seq),
        }

        self.model.p[self.part.name].SectionAssignment(**kwargs)



from . import abml_dataclass  # noqa
from . import abml_jnl_loader  # noqa
from . import abml_helpers  # noqa
from . import abml_materials  # noqa
from . import abml_sections  # noqa
from . import abml_parts  # noqa
from . import abml_sketch  # noqa
from . import abml_assembly  # noqa
from . import abml_mesh  # noqa
from . import abml_steps  # noqa
from . import abml_interaction_prop  # noqa
from . import abml_interactions  # noqa
from . import abml_constraints  # noqa
from . import abml_loads  # noqa
from . import abml_jobs  # noqa
from . import abml_bcs  # noqa



from abaqus import mdb
import sys
import os
import re


class IterReg(type):
    # Register of all objects for iteration
    def __iter__(cls):
        return iter(cls._reg.values())


class Abml_Helpers(object):
    """
    A utility class that provides helper methods for various tasks.

    Methods
    -------
    map_function(cls, key)
        A decorator method that maps the specified key to the provided function.
    get_boundary_rect(cls, sketch, depth, orientation)
        Calculates and returns the boundary rectangle of the given sketch with the given depth and orientation.
    convert_position_list(cls, position_list)
        Converts the provided position list into a standard list of lists format, if necessary.
    exit_handler()
        Cleans up temporary files created by the program.
    """

    @classmethod
    def map_function(cls, key):
        def decorator(func):
            cls.map[key] = func
            return func

        return decorator

    @classmethod
    def get_boundary_rect(cls, sketch, depth, orientation):
        p1 = sketch.kwargs["p1"]
        p2 = sketch.kwargs["p2"]
        x, y, z = abs(p1[0] - p2[0]), abs(p1[1] - p2[1]), depth

        offset = 1e-5
        boundaries = {
            "W": (-offset, 0, 0, 0, y, z),
            "N": (0, -offset, 0, x, 0, z),
            "T": (0, 0, -offset, x, y, 0),
            "top": (0, 0, -offset, x, y, 0),
            "E": (x - offset, 0, 0, x, y, z),
            "S": (0, y - offset, 0, x, y, z),
            "B": (0, 0, z - offset, x, y, z),
            "bot": (0, 0, z - offset, x, y, z),
        }

        return boundaries[orientation]

    @classmethod
    def convert_position_list(cls, position_list):
        """
        Convert the input position list to a list of lists.

        Parameters
        ----------
        position_list : list or None
            A list of positions to be converted.

        Returns
        -------
        position_list : list or None
            A list of lists of positions.

        Notes
        -----
        This method is used to convert a position list to a list of lists, where each sub-list
        contains a single position. If the input position list is already a list of lists, or if it
        is None, it is returned unmodified.

        Examples
        --------
        **python
        >>> Abml_Helpers.convert_position_list([1, 2, 3])
        [[1, 2, 3]]

        >>> Abml_Helpers.convert_position_list([[1, 2, 3], [4, 5, 6]])
        [[1, 2, 3], [4, 5, 6]]

        >>> Abml_Helpers.convert_position_list(None)
        None
        """

        if position_list is not None:

            def depth(list_):
                if isinstance(list_, list) and len(list_) != 0:
                    return 1 + max(depth(item) for item in list_)
                return 0

            depth_list = depth(position_list)
            if depth_list == 1 and all(isinstance(x, (int, float)) for x in position_list):
                position_list = [position_list]
            elif depth_list == 0:
                position_list = [position_list]
            return position_list
        return None


def exit_handler():
    """
    Deletes all files in the current directory with the extensions '.rpy', '.rec', and '.dmp' to clean up after program execution.
    """
    for file in os.listdir("."):
        if ".rpy" in file:
            try:
                os.remove(file)
            except WindowsError:
                pass
        if ".rec" in file:
            try:
                os.remove(file)
            except WindowsError:
                pass
        if ".dmp" in file:
            try:
                os.remove(file)
            except WindowsError:
                pass


def cprint(string, prefix="info"):
    """
    Prints the given string with a specified prefix to the standard output.

    Parameters
    ----------
    string : str
        The string to be printed.
    prefix : str, optional
        The prefix to be added to the string (default is "info").

    Returns
    -------
    None

    Notes
    -----
    This function uses the standard output stream `sys.__stdout__` to print the message.

    Examples
    --------
    **python
    >>> cprint("This is a message.", "warning")
    warning This is a message.

    >>> cprint("This is another message.")
    info This is another message.
    """
    print >> sys.__stdout__, prefix + " " + str(string)  # type: ignore # noqa


def _camel(match):
    """
    Helper function used in regex substitution to convert a matched string from camel case to snake case.

    Parameters:
    -----------
    match: MatchObject
        The matched object containing the string to be converted.

    Returns:
    --------
    str:
        The converted string in Pascal case.
    """
    return match.group(1) + match.group(2).upper()


patter_camel_to_snake = re.compile(r"(?<!^)(?=[A-Z])")


def camel_to_snake(name):
    """
    Convert a camelCase string to snake_case.

    Parameters
    ----------
    name : str
        The string to be converted.

    Returns
    -------
    str
        The converted string in snake_case.

    Examples
    --------
    **python
    >>> camel_to_snake("camelCaseString")
    'camel_case_string'
    >>> camel_to_snake("AnotherCamelCaseExample")
    'another_camel_case_example'
    """
    return patter_camel_to_snake.sub("_", name).lower()


patter_snake_to_camel = re.compile(r"(.*?)_([a-zA-Z])")


def snake_to_camel(name):
    """
    Convert snake_case to camelCase.

    Parameters
    ----------
    name : str
        The string to convert.

    Returns
    -------
    str
        The camelCase version of the input string.
    """
    return patter_snake_to_camel.sub(_camel, name, 0)


def get_objects(model_name):
    """
    Get the objects from a given modelname.

    Parameters
    ----------
    model_name : str
        The name of the model.

    Returns
    -------
    m : mdb.models[model_name]
        The Abaqus model object.
    a : mdb.models[model_name].assembly
        The Abaqus root assembly object.
    p : mdb.models[model_name].parts
        The Abaqus parts object.
    """

    m = mdb.models[model_name]
    a = m.rootAssembly
    p = m.parts

    return m, a, p


def get_kwargs_string(**kwargs):
    return " ".join(["{}:{}".format(key, value) for key, value in kwargs.items()])
